/**
  *  \file client/dialogs/racename.cc
  *  \brief Race Name Edit Dialog
  *
  *  (c) 2001-2019 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "client/dialogs/racename.h"
#include "ui/window.h"
#include "util/nls.h"
#include "game/racename.h"
#include "game/cmds.h"
#include "ui/widgets/inputline.h"
#include "gfx/palette.h"
#include "ui/layout.h"
#include "ui/uiglobal.h"
#include "ui/widgets/statictext.h"
#include "ui/widgets/focusiterator.h"
#include "ui/widgets/button.h"
#include "game/team.h"
#include "ui/widgets/quit.h"

static const char* TITLES[] = {
    N_("Full name:"),
    N_("Short name:"),
    N_("Adjective:"),
};
static const int LENGTHS[] = {
    GRaceNameList::FullNameSize,
    GRaceNameList::ShortNameSize,
    GRaceNameList::AdjNameSize,
};
static const int IDS[] = {
    GCommand::phcrn_Long,
    GCommand::phcrn_Short,
    GCommand::phcrn_Adjective,
};

static UIInputLine&
buildInput(ObjectHolder& h,
           UIGroup& group,
           int what,
           GCommandContainer& cmds,
           const string_t& full)
{
    const char* ttl = _(TITLES[what]);
    UIInputLine& input = h.add(new UIInputLine(LENGTHS[what], 17, UIKeyString(ttl).getKey(), UIInputLine::ilf_NoHi | UIInputLine::ilf_Framed, 1000 + what));

    group.add(h.add(new UIStaticText(ttl, UIStaticText::StaticTitle)));
    group.add(input);

    if (const GCommand* cmd = cmds.getCommand(GCommand::phc_SetRaceName, IDS[what])) {
        input.setText(cmd->getArg());
    } else {
        input.setText(full);
    }
    return input;
}

void
editRaceNames(GCommandContainer& cmds)
{
    // VBox
    //   Grid
    //     "Full name"    [edit]
    //     "Short name"   [edit]
    //     "Adjective"    [edit]
    //   UIStaticText "..."
    //   HBox
    //     "OK"
    //     "Cancel"

    ObjectHolder h;
    UIWindow& window = h.add(new UIWindow(_("Change Race Name"), UIWindow::BLUE_WINDOW, COLOR_GRAY));
    window.setLayout(UIVBoxLayout::instance5);
    UIGroup& group1 = h.add(new UIGroup(h.add(new UIGridLayout(2, 5))));
    UIGroup& group3 = h.add(new UIGroup(UIHBoxLayout::instance5));

    UIInputLine& inputFull  = buildInput(h, group1, 0, cmds, player_racenames.getFullName(getPlayerId()));
    UIInputLine& inputShort = buildInput(h, group1, 1, cmds, player_racenames.getShortName(getPlayerId()));
    UIInputLine& inputAdj   = buildInput(h, group1, 2, cmds, player_racenames.getAdjName(getPlayerId()));
    window.add(group1);
    window.add(h.add(new UIStaticText(_("Changes take effect next turn."), UIStaticText::StaticNormal)));

    group3.add(h.add(new UIButton(1, _("OK"), SDLK_RETURN)).addStop());
    group3.add(h.add(new UIButton(0, _("Cancel"), SDLK_ESCAPE)).addStop());
    group3.add(h.add(new UISpacer()));
    window.add(group3);
    window.add(h.add(new UIFocusIterator(1000, 1002, UIFocusIterator::fi_Vert | UIFocusIterator::fi_Tab)));
    window.add(h.add(new UIQuit(0)));

    window.pack();

    inputFull.setFocusedWidget();
    inputFull.setText(inputFull.getText()); // fix the cursor

    ui_root->centerWidget(window);
    if (ui_root->runModal(window)) {
        cmds.setCommand(GCommand::phc_SetRaceName, GCommand::phcrn_Long,      inputFull.getText());
        cmds.setCommand(GCommand::phc_SetRaceName, GCommand::phcrn_Short,     inputShort.getText());
        cmds.setCommand(GCommand::phc_SetRaceName, GCommand::phcrn_Adjective, inputAdj.getText());
    }
}
