/**
  *  \file client/screens/minescreen.cc
  *  \brief Minefield Screen
  *
  *  The Minefield screen currently just is a dialog.
  *
  *  (c) 2001-2019 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "client/screens/minescreen.h"
#include "client/screens.h"
#include "client/tiles/mineinfo.h"
#include "client/tiles/minepassage.h"
#include "client/tiles/objheader.h"
#include "client/widgets/objcontrol.h"
#include "cpluslib/format.h"
#include "game/config.h"
#include "game/global.h"
#include "game/host.h"
#include "game/mine.h"
#include "game/racename.h"
#include "game/team.h"
#include "gfx/palette.h"
#include "int/if/userif.h"
#include "ui/layout.h"
#include "ui/msgbox.h"
#include "ui/richdocument.h"
#include "ui/ui.h"
#include "ui/uiglobal.h"
#include "ui/widgets/button.h"
#include "ui/widgets/keytranslator.h"
#include "ui/window.h"
#include "util/misc.h"
#include "util/nls.h"
#include "client/widgets/objchart.h"
#include "client/widgets/helpwidget.h"
#include "ui/widgets/quit.h"
#include "int/context.h"

namespace {
    class WMinefieldInfoWindow : public UIWindow {
        ObjectHolder h;
        GObjectSelection& selection;
        void init();
     public:
        WMinefieldInfoWindow(GObjectSelection& selection);
        bool handleEvent(const UIEvent& event, bool second_pass);
    };
}

static RichText
boldIf(RichText in, bool flag)
{
    return flag
        ? in.withAttribute(RichTextStyleAttribute::Bold)
        : in;
}

static void
showSweepInfo(const GMinefield& mf)
{
    /* Figure out fighter sweep rate */
    int frace = ::getPlayerId();
    int frate = config.FighterSweepRate(frace);
    if (frate == 0) {
        /* We cannot sweep with fighters. Find someone who can. */
        for (int i = NUM_PLAYERS; i > 0; --i) {
            if (config.FighterSweepRate(i) > 0) {
                frate = config.FighterSweepRate(i);
                frace = i;
            }
        }
    }
    if (mf.isWeb() && (!config.AllowColoniesSweepWebs() || raceId(frace) != 11)) {
        /* This field cannot be swept with fighters */
        frate = 0;
    }

    /* Figure out units and regular sweep rate */
    int32_t units = host.isPHost() ? mf.getUnitsAfterDecay(mf.getUnits()) : mf.getUnits();
    int32_t rate = mf.isWeb() ? config.WebMineSweepRate(::getPlayerId()) : config.MineSweepRate(::getPlayerId());
    if (rate <= 0) {
        rate = 1;
    }

    /* Figure out available beam type */
    int beamType = 0;
    {
        ptr_vector<IntContext> ctxs;
        IntUserInterfaceBinding::get().enumContexts(ctxs);
        for (size_t i = ctxs.size(); i > 0; --i) {
            if (GShip* sh = dynamic_cast<GShip*>(ctxs[i-1]->getObject())) {
                if (sh->getBeamType().isKnown()) {
                    beamType = sh->getBeamType();
                }
                break;
            }
        }
    }

    /* Create dialog */
    int lines = 12 + (frate != 0);

    // VBox
    //   UIRichDocument
    //   HBox
    //     Spacer, "OK", Spacer
    ObjectHolder h;
    UIWindow& win = h.add(new UIWindow(_("Mine Sweep"), UIWindow::BLUE_WINDOW, COLOR_GRAY));
    win.setLayout(UIVBoxLayout::instance5);

    UIRichDocument& doc = h.add(new UIRichDocument(font_sizes[FONT_NORMAL].scaledBy(20, lines), 0));
    win.add(doc);

    UIGroup& g = h.add(new UIGroup(UIHBoxLayout::instance5));
    g.add(h.add(new UISpacer()));
    g.add(h.add(new UIButton(1, _("OK"), SDLK_RETURN)).addStop());
    g.add(h.add(new UISpacer()));
    win.add(g);
    win.add(h.add(new UIKeyTranslator()).addStop(' ', 1).addStop(SDLK_ESCAPE, 1));

    /* Generate the text */
    const int em = font_ems[FONT_NORMAL];
    RichDocument& d = doc.getDocument();
    d.setPageWidth(20*font_ems[FONT_NORMAL]);
    d.add(string_t(format(mf.isWeb() ? _("To sweep %d web mines, use...") : _("To sweep %d mines, use..."), numToString(units))));
    d.addParagraph();
    for (int i = 1; i <= NUM_BEAMS; ++i) {
        int32_t perBeam = i*i*rate;
        d.addRight(5*em, boldIf(numToString((units + perBeam-1) / perBeam), i == beamType));
        d.add(boldIf(string_t(" " + getBeam(i).getName()), i == beamType));
        d.addNewline();
    }
    if (frate != 0) {
        d.addRight(5*em, numToString((units + frate-1) / frate));
        d.add(string_t(format(_(" %s fighter"), player_racenames.getAdjName(frace))));
    }
    d.finish();

    /* Layout */
    doc.adjustToDocumentSize();
    win.pack();
    ui_root->centerWidget(win);

    /* Do it */
    ui_root->runModal(win);
}

/************************** WMinefieldInfoWindow *************************/

WMinefieldInfoWindow::WMinefieldInfoWindow(GObjectSelection& selection)
    : UIWindow(_("Minefield Information"), BLUE_WINDOW, COLOR_GRAY),
      selection(selection)
{
    init();
}

void
WMinefieldInfoWindow::init()
{
    // VBox
    //   WObjectSelectionHeaderTile
    //   HBox
    //     VBox
    //       WMinefieldInfoTile
    //       WMinefieldPassageTile
    //     WObjectSelectionChartWidget
    //   HBox
    //     Close GoTo Planet Del <--> H
    // Difference to PCC 1.x: Up/Down are in WObjectSelectionHeaderTile
    // FIXME: where are S and I?
    setLayout(UIVBoxLayout::instance5);

    UIGroup& g1 = h.add(new UIGroup(UIHBoxLayout::instance5));

    g1.add(h.add(new UIButton(0, _("Close"), SDLK_ESCAPE)).addStop());
    g1.add(h.add(new UIButton(1, _("Go To"), SDLK_RETURN)).addStop());
    g1.add(h.add(new UIButton(_("Del - Delete"), SDLK_DELETE)));
    g1.add(h.add(new UISpacer()));
    g1.add(h.add(new UIButton("H", 'h')));

    add(h.add(new WObjectSelectionHeaderTile(selection)));

    UIGroup& g2 = h.add(new UIGroup(UIHBoxLayout::instance5));
    UIGroup& g21 = h.add(new UIGroup(UIVBoxLayout::instance5));
    g21.add(h.add(new WMinefieldInfoTile(selection)));
    g21.add(h.add(new WMinefieldPassageTile(selection)));
    g2.add(g21);
    g2.add(h.add(new WObjectSelectionChartWidget(selection)));
    add(g2);
    add(g1);
    add(h.add(new WObjectSelectionControl(selection)));
    add(h.add(new WHelpWidget("pcc2:minescreen")));
    add(h.add(new UIQuit(0)));

    pack();
}

bool
WMinefieldInfoWindow::handleEvent(const UIEvent& event, bool second_pass)
{
    if (event.type == ev_Key) {
        if (event.key.code == 's') {
            if (GMinefield* mf = dynamic_cast<GMinefield*>(selection.getCurrentObject())) {
                showSweepInfo(*mf);
            }
            return true;
        }
        if (event.key.code == 'p') {
            GMinefield* mf = dynamic_cast<GMinefield*>(selection.getCurrentObject());
            GUniverse* univ = selection.getCurrentUniverse();
            if (mf != 0 && univ != 0) {
                int pid = getAssociatedPlanet(*univ, *mf);
                if (current_anyplanet_type.isValidIndex(pid)) {
                    stop(0);
                    postGoToObject(IntUserInterfaceBinding::get(), current_anyplanet_type.getObjectByIndex(pid), false);
                }
            }
            return true;
        }
        if (event.key.code == SDLK_DELETE) {
            if (UIMessageBox(_("Do you want to remove this minefield from the starchart? "
                               "Note that this will not sweep the field, you just won't "
                               "see it any longer."),
                             _("Delete Minefield")).doYesNoDialog())
            {
                GUniverse& univ = *selection.getCurrentUniverse();
                univ.ty_minefields.erase(selection.getCurrentIndex());
                univ.doScreenUpdates();
                if (selection.getObjectType()->isEmpty()) {
                    stop(0);
                }
            }
            return true;
        }
    }
    return UIWindow::handleEvent(event, second_pass);
}

void
doMinefieldScreen()
{
    if (current_minefield_type.isEmpty())
        return;

    WMinefieldInfoWindow win(current_minefield_selection);
    ui_root->centerWidget(win);
    int result = ui_root->runModal(win);
    if (result == 1) {
        // "Go to"
        if (GMapObject* mo = dynamic_cast<GMapObject*>(current_minefield_selection.getCurrentObject())) {
            postGoToChart(IntUserInterfaceBinding::get(), mo->getPos(), mo);
        }
    }
}
