/**
  *  \file game/fleet.cc
  *  \brief Fleets
  *
  *  Fleets are groups of ships with a common waypoint.
  *
  *  Fleet attributes are stored in GShip objects:
  *  - the fleet Id is the ship Id of the leader
  *  - the fleet has an optional name/comment which is stored in the leader
  *
  *  Fleet invariants:
  *  - if the leader has mission "Intercept X", all members have mission "Intercept X",
  *    and the same warp speed
  *    . if X is a member of that fleet, it has no waypoint and no speed.
  *  - otherwise, all ships have the same waypoint and the same warp speed.
  *    . as an exception, a ship that is being towed has no waypoint and no speed.
  *
  *  Loading:
  *  - loadFleets(): Fleet data is loaded from fleetX.cc. At this point, everything
  *    we have to make sure is that we do not stomp on other fleetX.cc which might
  *    be loaded in parallel, but we do not have access to derived information,
  *    just shipX.dat data and source flags.
  *  - postprocessFleet(): This is called after ship derived information has been
  *    set. At this point, we can assume that data is syntactically correct, but
  *    it could still happen that a fleet member might been gone. However, we can
  *    use the full API and data set here.
  *
  *  (c) 2001-2019 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include <cstring>
#include "game/fleet.h"
#include "cpluslib/format.h"
#include "game/config.h"
#include "game/global.h"
#include "game/mission.h"
#include "game/ship.h"
#include "game/struct.h"
#include "int/mutex.h"
#include "io/dirs.h"
#include "io/io.h"
#include "io/streams.h"
#include "util/except.h"
#include "util/nls.h"
#include "util/recode.h"

static const char MAGIC[] = { 'C', 'C', 'f', 'l', 'e', 'e', 't', 26 };
static const char FLEETFILE[] = "fleet%d.cc";

static const uint16_t Name_Flag = 0x8000;

/** Loading: Check whether ship is a valid fleet member for a given player. For use during construction.
    This checks source flags, not isPlayable(), because only source flags are valid at this time.
    \param sh Ship to check
    \param player Player to check for */
static bool
isValidFleetMember(const GShip& sh, const int player)
{
    return sh.getShipSource().contains(player) && sh.getOwner() == player && sh.getFleetNumber() == 0;
}

/** Loading: Build fleet from data. This takes fleet membership data loaded from the fleet file
    and places it into the universe, avoiding conflicts with existing fleets.
    \param univ     [in/out] Universe
    \param fid      [in] Fleet Id, corresponding to an existing fleet
    \param player   [in] Player number
    \param fleetNrs [in/out] Fleet numbers loaded from file, indexed [0,NUM_SHIPS). Used slots
                    will be zeroed.
    \param nameNrs  [in/out] Name pointers, indexed [0,NUM_SHIPS). A nonzero entry means to load
                    a name. Positive means to make this name the name of the corresponding fleet,
                    negative means to ignore the name. */
static void
buildFleet(GUniverse& univ, const uint16_t fid, const int player, uint16_t fleetNrs[], int16_t nameNrs[])
{
    // Find new fleet Id
    uint16_t newfid = 0;
    if (univ.isValidShipId(fid) && fleetNrs[fid-1] == fid && isValidFleetMember(univ.getShip(fid), player)) {
        // Keep fleet Id (also keep names)
        newfid = fid;
    } else {
        for (uint16_t i = 1; i <= NUM_SHIPS; ++i) {
            if (fleetNrs[i-1] == fid && isValidFleetMember(univ.getShip(i), player)) {
                // Found another ship that can get this fleet Id
                newfid = i;
                break;
            }
        }
    }

    // Find new name
    if (newfid == 0) {
        // Nothing found, fleet got completely annihilated, we have to delete it.
        if (fid > 0 && fid <= NUM_SHIPS && nameNrs[fid-1] != 0) {
            nameNrs[fid-1] = -1;
        }
    } else if (newfid != fid) {
        // Id changed, move name:
        // - Place fid's name in newfid.
        if (fid > 0 && fid <= NUM_SHIPS && nameNrs[fid-1] != 0) {
            nameNrs[fid-1] = newfid;
        }
        // - Discard newfid's name.
        if (nameNrs[newfid-1] != 0) {
            nameNrs[newfid-1] = -1;
        }
    } else {
        // Id remains the same.
    }

    // Build fleet and strike out of fleetNrs
    if (newfid != 0) {
        ASSERT(isValidFleetMember(univ.getShip(newfid), player));
    }
    for (uint16_t i = 1; i <= NUM_SHIPS; ++i) {
        if (fleetNrs[i-1] == fid) {
            GShip& sh = univ.getShip(i);
            if (isValidFleetMember(sh, player) || newfid == 0) {
                sh.setFleetNumber(newfid);
                fleetNrs[i-1] = 0;
            }
        }
    }
}

/** Manipulating: Remove member from a fleet.
    \param univ Universe to work on
    \param sid Ship Id
    \return Id of fleet this ship was removed from (new Id if fleet Id changed) */
static int
removeFleetMember(GUniverse& univ, const int sid)
{
    GShip& sh = univ.getShip(sid);
    const int fid = sh.getFleetNumber();

    if (fid != sid) {
        // Member
        sh.setFleetNumber(0);
        return fid;
    } else {
        // Leader. This means we have to find a new leader.
        int newfid = 0;
        for (int i = 1; i <= NUM_SHIPS; ++i) {
            if (i != sid && univ.getShip(i).isPlayable(GObject::Playable) && univ.getShip(i).getFleetNumber() == fid) {
                newfid = i;
                break;
            }
        }
        if (newfid != 0) {
            univ.getShip(newfid).setFleetName(sh.getFleetName());
            univ.getShip(newfid).setFleetNumber(newfid);
        }
        sh.setFleetName(string_t());
        sh.setFleetNumber(0);

        // Move all members. This does not check isPlayable(), thus also moves nonexistant members,
        // which are cleaned up by postprocessFleet().
        for (int i = 1; i <= NUM_SHIPS; ++i) {
            if (univ.getShip(i).getFleetNumber() == fid) {
                univ.getShip(i).setFleetNumber(newfid);
            }
        }
        return newfid;
    }
}

/** Manipulating: Postprocess a fleet. This will remove all fleet members that do not
    exist anymore. It will also synchronize the waypoints. */
static void
postprocessFleet(GUniverse& univ, int fid)
{
    // Check leader
    if (!univ.isValidShipId(fid) || !univ.getShip(fid).isPlayable(GObject::Playable)) {
        fid = removeFleetMember(univ, fid);
    }

    // Check members
    if (fid != 0) {
        for (int i = 1; i <= NUM_SHIPS; ++i) {
            if (univ.getShip(i).getFleetNumber() == fid && !univ.getShip(i).isPlayable(GObject::Playable)) {
                removeFleetMember(univ, i);
            }
        }
    }

    // Synchronize
    synchronizeFleet(univ, fid);
}

/** Manipulating: check whether fleet member is being towed. */
static bool
isFleetMemberTowed(const GUniverse& univ, const int sid)
{
    const int fid = univ.getShip(sid).getFleetNumber();
    if (fid != 0) {
        for (int i = 1; i <= NUM_SHIPS; ++i) {
            const GShip& sh = univ.getShip(i);
            if (i != sid && sh.getFleetNumber() == fid) {
                if (sh.getMission() == GMission::msn_Tow && sh.getTowId() == sid) {
                    return true;
                }
            }
        }
    }

    return false;
}

/** Manipulating: set waypoint for intercepting ship. */
static void
setInterceptWaypoint(GUniverse& univ, const int sid)
{
    GShip& sh = univ.getShip(sid);
    const int i = sh.getInterceptId();
    if (univ.ty_any_ships.isValidIndex(i)) {
        // FIXME: handle THost where intercept does not cross the seam?
        sh.setWaypoint(univ.getShip(i).getPos().getSimpleNearestAlias(sh.getPos()));
    }
}

/** Process towee of a ship. */
static void
synchronizeTowee(GUniverse& univ, const int sid)
{
    GShip& ship = univ.getShip(sid);
    if (ship.getMission() == GMission::msn_Tow) {
        int tow = ship.getTowId();
        if (univ.ty_played_ships.isValidIndex(tow)) {
            synchronizeFleetMember(univ, tow);
        }
    }
}

/******************************* GFleetType ******************************/

GFleetType::GFleetType(GUniverse& univ)
    : GIndexedObjectType(NUM_SHIPS),
      univ(univ)
{ }

GShip&
GFleetType::getObjectByIndex(int index)
{
    return univ.getShip(index);
}

GUniverse*
GFleetType::getUniverseByIndex(int /*index*/)
{
    return &univ;
}

bool
GFleetType::isValidIndex(int index) const
{
    if (index > 0 && index <= NUM_SHIPS) {
        const GShip& s = univ.getShip(index);
        return (s.isPlayable(GObject::Playable) && s.getFleetNumber() == index);
    } else {
        return false;
    }
}

/** Handle fleet change. If a change caused the current fleet to get invalid,
    this finds a new one.
    \param hintfid Hint for a possible fleet Id. May or may not be valid. */
void
GFleetType::handleFleetChange(const int hintfid)
{
    sig_set_changed.raise(hintfid);
}

/**************************** Public Functions ***************************/

/** Load fleets. This will load the given player's fleetX.cc file, and
    populate his ships with the fleet data. This validates the data as far
    as required to make sure there are no collisions. */
void
loadFleets(GGameTurn& trn, Directory& dir, const int player)
{
    GUniverse& univ = trn.getCurrentUniverse();

    // Open file
    Ptr<Stream> s = dir.openFileNT(format(FLEETFILE, player), Stream::C_READ);
    if (!s) {
        return;
    }

    // Read header
    char header[9];
    s->readT(header, sizeof(header));
    if (std::memcmp(header, MAGIC, 8) != 0 || (header[8] != 0 && header[8] != 1)) {
        throw FileFormatException(*s, _("Invalid file header"));
    }

    // Get count
    uint16_t nFleets = 500;
    if (header[8] == 1) {
        TInt16 count;
        getStructureT(*s, count);
        nFleets = count.value;
        if (nFleets > NUM_SHIPS) {
            throw FileFormatException(*s, _("Invalid count"));
        }
    }

    // Get bits
    char fleetData[NUM_SHIPS*2];
    s->readT(fleetData, 2*nFleets);

    // Extract fleet numbers.
    uint16_t fleetNrs[NUM_SHIPS];
    getUint16(fleetData, fleetNrs, nFleets);
    for (uint16_t i = nFleets; i < NUM_SHIPS; ++i) {
        fleetNrs[i] = 0;
    }

    // Prepare names. nameNrs is nonzero to load a name at this place,
    // positive to load it into that ship.
    int16_t nameNrs[NUM_SHIPS];
    for (uint16_t i = 0; i < NUM_SHIPS; ++i) {
        if (fleetNrs[i] & Name_Flag) {
            fleetNrs[i] &= ~Name_Flag;
            if (fleetNrs[i] == i+1) {
                // Fleet leader
                nameNrs[i] = i+1;
            } else {
                // Fleet member (comment invalid here)
                nameNrs[i] = -1;
            }
        } else {
            // No comment
            nameNrs[i] = 0;
        }
    }

    // Postprocess the fleets
    for (uint16_t i = 0; i < NUM_SHIPS; ++i) {
        if (uint16_t nr = fleetNrs[i]) {
            buildFleet(univ, nr, player, fleetNrs, nameNrs);
        }
    }

    // Load comments
    for (uint16_t i = 0; i < NUM_SHIPS; ++i) {
        string_t comment;
        if (nameNrs[i] != 0) {
            getPascalString(*s, comment);
            if (nameNrs[i] > 0) {
                ASSERT(univ.isValidShipId(nameNrs[i]));
                ASSERT(univ.getShip(nameNrs[i]).getFleetNumber() == nameNrs[i]);
                univ.getShip(nameNrs[i]).setFleetName(convertGameToUtf8(comment));
            }
        }
    }
}

/** Save fleets. Saves the given player's fleets. */
void
saveFleets(const GGameTurn& trn, Directory& dir, const int player)
{
    // Build fleet list and find highest fleet member
    uint16_t highestFleetMember = 0;
    uint16_t fleetNrs[NUM_SHIPS];
    for (uint16_t i = 1; i <= NUM_SHIPS; ++i) {
        const GShip& sh = trn.getCurrentUniverse().getShip(i);
        fleetNrs[i-1] = 0;
        if (sh.getShipSource().contains(player) && sh.getOwner() == player) {
            int nr = sh.getFleetNumber();
            if (nr != 0) {
                fleetNrs[i-1] = nr;
                if (nr == i && !sh.getFleetName().empty()) {
                    fleetNrs[i-1] |= Name_Flag;
                }
                highestFleetMember = i;
            }
        }
    }

    // If there is no fleet, erase the fleet file
    const string_t fname = format(FLEETFILE, player);
    if (highestFleetMember == 0) {
        dir.eraseNT(fname);
        return;
    }

    // Decide upon file format (Host999 or normal)
    char header[11];
    uint16_t headerSize, nFleets;
    std::memcpy(header, MAGIC, 8);
    if (highestFleetMember <= 500) {
        header[8] = 0;
        headerSize = 9;
        nFleets = 500;
    } else {
        header[8] = 1;
        headerSize = 11;
        nFleets = 999;
        storeUint16(&header[9], nFleets);
    }

    // Convert
    char fleetData[NUM_SHIPS*2];
    storeUint16(fleetData, fleetNrs, nFleets);

    // Write it
    Ptr<Stream> s = dir.openFile(fname, Stream::C_CREATE);
    s->writeT(header, headerSize);
    s->writeT(fleetData, nFleets*2);

    // Write comments
    for (uint16_t i = 1; i <= NUM_SHIPS; ++i) {
        if (fleetNrs[i-1] & Name_Flag) {
            storePascalStringT(*s, convertUtf8ToGame(trn.getCurrentUniverse().getShip(i).getFleetName()));
        }
    }
}

/** Postprocess all fleets. This assumes that fleet data is syntactically correct,
    i.e. all fleet flags are consistent, but there may be errors such as nonexistant
    fleet members. It also synchronizes all fleet waypoints. */
void
postprocessFleets(GGameTurn& trn)
{
    GUniverse& univ = trn.getCurrentUniverse();
    for (int i = univ.ty_any_ships.findNextIndex(0); i != 0; i = univ.ty_any_ships.findNextIndex(i)) {
        if (univ.getShip(i).getFleetNumber() == i) {
            // Note that postprocessFleet may change the fleet Id.
            // The fleet may thus be processed twice, which is harmless.
            postprocessFleet(univ, i);
        }
    }
}

/** Leave a fleet. Makes the specified ship leave the fleet. Handles all cases
    (in particular, the case of the leader leaving the fleet).
    \param univ Universe to work on
    \param sid Ship Id */
void
leaveFleet(GUniverse& univ, const int sid)
{
    const int oldfid = univ.getShip(sid).getFleetNumber();
    if (oldfid != 0) {
        // Remove
        const int fid = removeFleetMember(univ, sid);
        univ.ty_fleets.handleFleetChange(fid);

        // If we are towing a fleet member, they may now get free
        synchronizeTowee(univ, sid);
        
        // Notify
        univ.getShip(sid).markDirty();
        if (fid != 0) {
            if (fid != oldfid) {
                // Fleet Id changed, so notify whole fleet
                markFleetDirty(univ, fid);
            } else {
                // Notify just the leader
                univ.getShip(fid).markDirty();
            }
        }
    }
}

/** Synchronize a fleet member. Synchronizes waypoint of a single fleet member.
    \param univ Universe to work on
    \param sid Ship Id */
void
synchronizeFleetMember(GUniverse& univ, const int sid)
{
    // Fetch ship
    GShip& sh = univ.getShip(sid);
    const int fid = sh.getFleetNumber();
    if (fid == 0) {
        return;
    }

    if (sid == fid) {
        // I am the leader, so nothing to do.
    } else {
        // I am a member. Fetch the leader.
        const GShip& leader = univ.getShip(fid);

        // Check for intercept mission
        const GMission*const msn = GMissionList::getInstance().getMissionByShip(leader);
        const bool xcept = (msn != 0 && (msn->getFlags() & GMission::mf_Waypoint) != 0);

        if (xcept) {
            // We're intercepting
            if (leader.getInterceptId() == sid) {
                // Fleet is intercepting this very ship.
                sh.setMission(mp16_t(GMission::msn_Explore), mp16_t(0), mp16_t(0));
                sh.setWaypoint(sh.getPos());
                sh.setWarp(mp16_t(getEngine(sh.getEngineType()).getMaxEfficientWarp()));
            } else {
                // Fleet is intercepting someone else.
                sh.setMission(leader.getMission(), leader.getInterceptId(), leader.getTowId());
                setInterceptWaypoint(univ, sid);
                sh.setWarp(mp16_t(leader.getWarp()));
            }
        } else {
            // We're moving normally
            if (isFleetMemberTowed(univ, sid)) {
                // Member is under tow, so avoid that it escapes.
                sh.setWarp(mp16_t(0));
                sh.setWaypoint(sh.getPos());
            } else {
                // Normal move
                sh.setWarp(leader.getWarp());
                sh.setWaypoint(leader.getWaypoint().getSimpleNearestAlias(sh.getPos()));
            }

            // Cancel any intercept
            const GMission*const shmsn = GMissionList::getInstance().getMissionByShip(sh);
            if (shmsn != 0 && (shmsn->getFlags() & GMission::mf_Waypoint) != 0) {
                sh.setMission(mp16_t(GMission::msn_Explore), mp16_t(0), mp16_t(0));
            }
        }
    }        
}

/** Synchronize a fleet. Synchronizes all waypoints of the fleet.
    \param univ Universe to work on
    \param fid Fleet Id */
void
synchronizeFleet(GUniverse& univ, const int fid)
{
    for (uint16_t i = 1; i <= NUM_SHIPS; ++i) {
        GShip& sh = univ.getShip(i);
        if (sh.getFleetNumber() == fid) {
            synchronizeFleetMember(univ, i);
        }
    }
}

/** Mark a fleet dirty. Marks all fleet members dirty.
    \param univ Universe to work on
    \param fid Fleet Id */
void
markFleetDirty(GUniverse& univ, const int fid)
{
    for (uint16_t i = 1; i <= NUM_SHIPS; ++i) {
        GShip& sh = univ.getShip(i);
        if (sh.getFleetNumber() == fid) {
            sh.markDirty();
        }
    }
}

/** Set fleet number.
    \param univ Universe to work on
    \param sid Ship Id whose fleet number to change, must be existing ship
    \param nfid New fleet Id
    \retval true success
    \retval false error, nfid was not a valid fleet Id */
bool
setFleetNumber(GUniverse& univ, const int sid, const int nfid)
{
    GShip& sh = univ.getShip(sid);
    const int fid = sh.getFleetNumber();
    if (nfid == fid) {
        // no change, ok
        return true;
    } else if (nfid == 0) {
        // leave fleet
        leaveFleet(univ, sid);
        return true;
    } else if (univ.ty_played_ships.isValidIndex(nfid)) {
        GShip& nsh = univ.getShip(nfid);
        if (nsh.getOwner() == sh.getOwner()) {
            // enter or change fleet
            if (nfid == sid) {
                // start new fleet
                removeFleetMember(univ, sid);
                markFleetDirty(univ, fid);
                sh.setFleetNumber(nfid);
                univ.ty_fleets.handleFleetChange(nfid);
                sh.markDirty();
                synchronizeTowee(univ, sid);
                return true;
            } else if (nsh.getFleetNumber() == nfid) {
                // join a fleet
                removeFleetMember(univ, sid);
                markFleetDirty(univ, fid);
                sh.setFleetNumber(nfid);
                synchronizeFleetMember(univ, sid);
                markFleetDirty(univ, nfid);
                univ.ty_fleets.handleFleetChange(nfid);
                sh.markDirty();
                nsh.markDirty();
                synchronizeTowee(univ, sid);
                return true;
            } else {
                // invalid: not a fleet leader
                return false;
            }
        } else {
            // invalid
            return false;
        }
    } else {
        // invalid
        return false;
    }
}

/** Set name of a fleet.
    \param univ Universe to work on
    \param sid Ship or Fleet Id, must be Id of existing ship
    \param nname New name
    \retval true success
    \retval false error, %sid was not a valid fleet id */
bool
setFleetName(GUniverse& univ, const int sid, const string_t nname)
{
    GShip& sh = univ.getShip(sid);
    if (sh.getFleetNumber() == sid) {
        sh.setFleetName(nname);
        markFleetDirty(univ, sid);
        return true;
    } else {
        return false;
    }
}

/** Set waypoint of a fleet.
    \param univ Universe to work on
    \param sid Ship Id
    \param pt Desired waypoint
    \retval true success, waypoint of fleet or lone ship changed
    \retval false failure, this was a fleet member */
bool
setFleetWaypoint(GUniverse& univ, const int sid, const GPoint pt)
{
    GShip& sh = univ.getShip(sid);
    if (sh.isFleetMember()) {
        return false;
    } else {
        if (pt != sh.getWaypoint()) {
            // Set waypoint
            sh.setWaypoint(pt.getSimpleNearestAlias(sh.getPos()));

            // Cancel intercept, if any
            const GMission*const msn = GMissionList::getInstance().getMissionByShip(sh);
            if (msn != 0 && (msn->hasFlag(GMission::mf_Waypoint))) {
                sh.setMission(mp16_t(GMission::msn_Explore), mp16_t(0), mp16_t(0));
            }

            // Distribute change
            if (sh.isFleetLeader()) {
                synchronizeFleet(univ, sid);
            }
        }
        return true;
    }
}

/** Set speed of a fleet.
    \param univ Universe to work on
    \param sid Ship Id
    \param speed Desired speed
    \retval true success, speed of fleet or lone ship changed
    \retval false failure, this was a fleet member */
bool
setFleetSpeed(GUniverse& univ, const int sid, const int speed)
{
    GShip& sh = univ.getShip(sid);
    if (sh.isFleetMember()) {
        return false;
    } else {
        sh.setWarp(mp16_t(speed));
        if (sh.isFleetLeader()) {
            synchronizeFleet(univ, sid);
        }
        return true;
    }
}

/** Set mission of a fleet or member.
    \param univ Universe to work on
    \param sid Ship Id
    \param m,i,t Desired mission
    \retval true success, speed of fleet, member, or lone ship changed
    \retval false failure, this was a fleet member and this fleet was intercepting */
bool
setFleetMission(GUniverse& univ, const int sid, const int m, const int i, const int t)
{
    GShip& sh = univ.getShip(sid);

    // Is this a change after all?
    if (m == sh.getMission() && i == sh.getInterceptId() && t == sh.getTowId()) {
        return true;
    }

    // New mission
    const GMission* msn = GMissionList::getInstance().getMissionByNumber(m, GPlayerSet(sh.getRealOwner()));
    const bool xcept = (msn != 0 && msn->hasFlag(GMission::mf_Waypoint));

    // Check for intercept
    if (sh.isFleetMember()) {
        // Fail if old leader has intercept, or new mission is intercept.
        // Those are not allowed on members.
        if (xcept) {
            return false;
        }

        msn = GMissionList::getInstance().getMissionByShip(univ.getShip(sh.getFleetNumber()));
        if (msn != 0 && msn->hasFlag(GMission::mf_Waypoint)) {
            return false;
        }
    }

    // Remember old mission
    int oldMission = sh.getMission();
    int oldTow = sh.getTowId();

    // Valid change
    sh.setMission(mp16_t(m), mp16_t(i), mp16_t(t));

    // Postprocess intercept
    if (xcept) {
        // Set waypoint to intercept target
        setInterceptWaypoint(univ, sid);
    }

    // Propagate to members
    if (sh.isFleetLeader()) {
        synchronizeFleet(univ, sh.getId());
    }

    // Postprocess tow
    if (oldMission == GMission::msn_Tow) {
        if (univ.isValidShipId(oldTow) && univ.getShip(oldTow).getFleetNumber() == sh.getFleetNumber()) {
            synchronizeFleetMember(univ, oldTow);
        }
    }
    if (m == GMission::msn_Tow) {
        if (univ.isValidShipId(t) && univ.getShip(t).getFleetNumber() == sh.getFleetNumber()) {
            synchronizeFleetMember(univ, t);
        }
    }

    return true;
}

/** Check whether the fleet can do a particular special function.
    \param univ Universe to work on
    \param fid Fleet Id
    \param basic_function Function to check for
    \retval true all fleet members can do this function
    \retval false not all fleet members can do this function */
bool
canFleetDoSpecial(const GUniverse& univ, int fid, int basic_function)
{
    if (univ.getShip(fid).getFleetNumber() == 0) {
        // Lone ship: just check it
        return univ.getShip(fid).canDoSpecial(basic_function);
    } else {
        // Fleet: check all members
        for (uint16_t i = 1; i <= NUM_SHIPS; ++i) {
            const GShip& sh = univ.getShip(i);
            if (sh.getFleetNumber() == fid) {
                if (!sh.canDoSpecial(basic_function)) {
                    return false;
                }
            }
        }
        return true;
    }
}

/** Get title for a fleet. */
string_t
getFleetTitle(const GUniverse& univ, const int fid)
{
    const GShip& sh = univ.getShip(fid);
    if (sh.getFleetName().empty()) {
        return format(_("Fleet %d: led by %s"), fid, sh.getName(GObject::PlainName));
    } else {
        return format(_("Fleet %d: %s"), fid, sh.getFleetName());
    }
}

/** Check whether a mission is locked.
    \param sh Ship to check
    \param flags Modifiers
    \retval true this mission is locked and cannot be modified
    \retval false this mission is not locked */
bool
isMissionLocked(const GShip& sh, const int flags)
{
    /* Is it an intercept mission after all? */
    const GMission*const m = GMissionList::getInstance().getMissionByShip(sh);
    if (m == 0 || !m->hasFlag(GMission::mf_Waypoint)) {
        return false;
    }

    /* Is it locked because we're a fleet member? */
    if (sh.getFleetNumber() != 0) {
        if (!sh.isFleetLeader() || (flags & mlAcceptLeaders) == 0) {
            return true;
        }
    }

    /* Is it locked because someone has the waypoint mutex? */
    if ((flags & mlOverrideLocks) == 0 && IntMutex::query(format("S%d.WAYPOINT", sh.getId())) != 0) {
        return true;
    }

    /* Not locked */
    return false;
}
