/**
  *  \file gfx/bmfont.cc
  *  \brief Bitmap Fonts
  *
  *  (c) 2001-2019 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include <memory>
#include "gfx/bmfont.h"
#include "io/streams.h"
#include "cpluslib/buffer.h"
#include "util/nls.h"
#include "util/except.h"
#include "io/io.h"
#include "cpluslib/utf8.h"

static inline uint16_t
mapCharacterId(int /*encoding*/, uint16_t chid)
{
    /* We would remap character Ids here:
       encoding = 0(cp437/pcc1), 1(cp866), 2(unicode) */
    return chid;
}

/***************************** GfxBitmapGlyph ****************************/

/** Construct blank glyph. */
GfxBitmapGlyph::GfxBitmapGlyph()
    : width(0), height(0), data(), aadata()
{ }

/** Construct blank glyph.
    \param width  [in] Width in pixels
    \param height [in] Height in pixels */
GfxBitmapGlyph::GfxBitmapGlyph(uint16_t width, uint16_t height)
    : width(width),
      height(height),
      data(getBytesForSize(width, height)),
      aadata()
{ }

/** Construct glyph from input data.
    \param width  [in] Width in pixels
    \param height [in] Height in pixels
    \param data   [in] Refers to getBytesForSize(width,height) bytes containing bitmap data */
GfxBitmapGlyph::GfxBitmapGlyph(uint16_t width, uint16_t height, const uint8_t* data)
    : width(width),
      height(height),
      data(data, data + getBytesForSize(width, height)),
      aadata()
{ }

/** Destructor. */
GfxBitmapGlyph::~GfxBitmapGlyph()
{ }

/** Get number of bytes per line. */
inline int
GfxBitmapGlyph::getBytesPerLine() const
{
    return (width+7)/8;
}

/** Anti-aliasing hint. Specifies that the pixel at (x,y) should be
    drawn in half intensity.
    \param x X-coordinate [0,width)
    \param y Y-coordinate [0,height) */
void
GfxBitmapGlyph::addAAHint(uint16_t x, uint16_t y)
{
    aadata.push_back(x);
    aadata.push_back(y);
    set(x, y, 0);
}

/** Draw this glyph.
    \param ctx Context to draw on
    \param x,y Position */
void
GfxBitmapGlyph::draw(GfxContext& ctx, GfxPoint pt) const
{
    // Do nothing if this glyph is blank
    if (width == 0 || height == 0)
        return;

    // Main data
    ctx.getCanvas().blitPattern(GfxRect(pt, width, height),
                                pt,
                                getBytesPerLine(),
                                &*data.begin(),
                                ctx.getRawColor(),
                                GFX_TRANSPARENT,
                                ctx.getAlpha());

    // AA hints
    const alpha_type half_intensity = (ctx.getAlpha()+1)/2;
    for (std::vector<uint16_t>::size_type i = 0; i < aadata.size(); i += 2) {
        ctx.getCanvas().drawPixel(GfxPoint(pt.x + aadata[i], pt.y + aadata[i+1]), ctx.getRawColor(), half_intensity);
    }
}

/** Draw this glyph, with solid colors.
    \param can Canvas to draw on
    \param x,y Position
    \param pixel_color Color of regular pixels
    \param aa_color Color of half-intensity pixels */
void
GfxBitmapGlyph::drawColored(GfxCanvas& can, GfxPoint pt, uint32_t pixel_color, uint32_t aa_color) const
{
    // Do nothing if this glyph is blank
    if (width == 0 || height == 0)
        return;

    // Main data
    can.blitPattern(GfxRect(pt, width, height),
                    pt,
                    getBytesPerLine(),
                    &*data.begin(),
                    pixel_color,
                    GFX_TRANSPARENT,
                    GFX_ALPHA_OPAQUE);

    // AA hints
    for (std::vector<uint16_t>::size_type i = 0; i < aadata.size(); i += 2) {
        can.drawPixel(GfxPoint(pt.x + aadata[i], pt.y + aadata[i+1]), aa_color, GFX_ALPHA_OPAQUE);
    }
}


/** Set pixel value.
    \param x,y   Position
    \param value true to set pixel, false to clear it */
void
GfxBitmapGlyph::set(int x, int y, bool value)
{
    if (value) {
        data[y * getBytesPerLine() + (x >> 3)] |=  (0x80 >> (x&7));
    } else {
        data[y * getBytesPerLine() + (x >> 3)] &= ~(0x80 >> (x&7));
    }
}

/** Get pixel value.
    \param x,y   Position */
bool
GfxBitmapGlyph::get(int x, int y) const
{
    return (data[y * getBytesPerLine() + (x >> 3)] & (0x80 >> (x&7))) != 0;
}


/** Compute number of bytes required for a glyph of the specified size.
    \param width Width in pixels
    \param height Height in pixels */
uint32_t
GfxBitmapGlyph::getBytesForSize(uint16_t width, uint16_t height)
{
    return height * ((width + 7) / 8);
}

/***************************** GfxBitmapFont *****************************/

/** Constructor. Makes an empty font. */
GfxBitmapFont::GfxBitmapFont()
    : height(0)
{ }

/** Destructor. */
GfxBitmapFont::~GfxBitmapFont()
{ }

/** Add new glyph. GfxBitmapFont assumes ownership of the glyph.
    If there already is a glyph with that Id, it is replaced. */
void
GfxBitmapFont::addNewGlyph(uint16_t id, GfxBitmapGlyph* g)
{
    std::size_t outerIndex = id >> 8;
    std::size_t innerIndex = (id & 255);

    /* Do not enlarge arrays to store null in it. */
    if (g == 0
        && (outerIndex >= glyphs.size()
            || glyphs[outerIndex] == 0
            || innerIndex >= glyphs[outerIndex]->size()))
    {
        return;
    }

    try {
        /* Make sure outer array is big enough */
        while (glyphs.size() <= outerIndex)
            glyphs.push_back(0);

        /* Make sure inner array exists */
        if (glyphs[outerIndex] == 0)
            glyphs.replace_elem(outerIndex, new ptr_vector<GfxBitmapGlyph>());

        /* Make sure inner array is big enough */
        while (glyphs[outerIndex]->size() <= innerIndex)
            glyphs[outerIndex]->push_back(0);
    }
    catch (...) {
        delete g;
        throw;
    }

    /* Store glyph */
    glyphs[outerIndex]->replace_elem(innerIndex, g);

    /* Update height */
    if (g != 0) {
        int h = g->getHeight();
        if (h > height)
            height = h;
    }
}

/** Get glyph for a character.
    \param id Character id
    \return Glyph; 0 if none present. */
const GfxBitmapGlyph*
GfxBitmapFont::getGlyph(uint16_t id) const
{
    std::size_t outerIndex = id >> 8;
    std::size_t innerIndex = (id & 255);
    if (outerIndex < glyphs.size()
        && glyphs[outerIndex] != 0
        && innerIndex < glyphs[outerIndex]->size())
    {
        return (*glyphs[outerIndex])[innerIndex];
    } else {
        return 0;
    }
}

/** Get current upper bound of character Ids.
    This value can be used to control iteration across all characters
    ('for (uint32_t i = 0, e = getCurrentCharacterLimit(); i < e; ++i)').
    \return some n, such that getGlyph(m) returns null for all m >= n. */
uint32_t
GfxBitmapFont::getCurrentCharacterLimit() const
{
    uint32_t limit = 256 * glyphs.size();
    if (glyphs.size() != 0 && glyphs.back() != 0) {
        limit -= 256;
        limit += glyphs.back()->size();
    }
    return limit;
}


// GfxFont virtuals:
void
GfxBitmapFont::outText(GfxContext& ctx, GfxPoint pt, string_t text)
{
    Utf8Reader rdr(text);
    while (rdr.hasMore()) {
        uint32_t ch = rdr.get();
        if (const GfxBitmapGlyph* g = getGlyph(ch)) {
            // Render this character
            g->draw(ctx, pt);
            pt.x += g->getWidth();
        } else if (Utf8Reader::isErrorCharacter(ch)) {
            // Try to render two digits
            uint8_t base = ch - Utf8Reader::makeErrorCharacter(0x80) + 0x80;
            const GfxBitmapGlyph* a = getGlyph(0xE100 + ((base >> 4) & 15));
            const GfxBitmapGlyph* b = getGlyph(0xE130 + (base & 15));
            if (a) {
                a->draw(ctx, pt);
                if (b)
                    b->draw(ctx, pt);
                pt.x += a->getWidth();
            }
        } else {
            // Try to render four digits
            const GfxBitmapGlyph* a = getGlyph(0xE100 + ((ch >> 12) & 15));
            const GfxBitmapGlyph* b = getGlyph(0xE110 + ((ch >> 8) & 15));
            const GfxBitmapGlyph* c = getGlyph(0xE120 + ((ch >> 4) & 15));
            const GfxBitmapGlyph* d = getGlyph(0xE130 + (ch & 15));
            if (a) {
                a->draw(ctx, pt);
                if (b)
                    b->draw(ctx, pt);
                if (c)
                    c->draw(ctx, pt);
                if (d)
                    d->draw(ctx, pt);
                pt.x += a->getWidth();
            }
        }
    }
}

int
GfxBitmapFont::getTextWidth(string_t text)
{
    Utf8Reader rdr(text);
    int total = 0;
    while (rdr.hasMore()) {
        uint32_t ch = rdr.get();
        if (const GfxBitmapGlyph* g = getGlyph(ch)) {
            total += g->getWidth();
        } else if (Utf8Reader::isErrorCharacter(ch)) {
            if (const GfxBitmapGlyph* g = getGlyph(0xE100 + ((((ch - Utf8Reader::makeErrorCharacter(0x80)) + 0x80) >> 4) & 15)))
                total += g->getWidth();
        } else {
            if (const GfxBitmapGlyph* g = getGlyph(0xE100 + ((ch >> 12) & 15)))
                total += g->getWidth();
        }
    }
    return total;
}

int
GfxBitmapFont::getTextHeight(string_t /*text*/)
{
    return height;
}

bool
GfxBitmapFont::setScale(int mul, int div)
{
    return mul==div;
}

GfxPoint
GfxBitmapFont::getScale()
{
    return GfxPoint(1, 1);
}

unsigned
GfxBitmapFont::getCapabilities()
{
    return FN_LINEAR | FN_ALPHA;
}

/** Load bitmap font from "FN" file. This is a custom font file
    format, used by PCC 1.x as well as PCC2. */
void
GfxBitmapFont::load(Stream& s, int index)
{
    Stream::fpos_t base = s.getPos();

    // Read file header
    char header[16];
    s.readT(header, 4);
    if (header[0] != 'F' || header[1] != 'N')
        throw FileFormatException(s, _("File is missing required signature"));

    const int nfonts = getInt16(header+2);
    if (nfonts <= 0 || index >= nfonts)
        throw FileFormatException(s, _("File does not contain required font"));

    // Read font index
    s.seek(base + 4 + 8*index);
    s.readT(header, 8);
    const uint32_t pos      = getUint32(header);
    //const uint8_t flags   = header[4];
    const uint8_t  encoding = header[5];

    // Read font header
    s.seek(base + pos);
    s.readT(header, 6);
    const uint16_t type   = getUint16(header);
    const uint16_t height = getUint16(header+2);
    const uint16_t nchars = getUint16(header+4);

    // Read character headers
    const uint32_t chsize = nchars*8;
    Array<char> tmp(chsize);
    s.readT(tmp.ptr(), chsize);

    // Read characters
    for (uint32_t i = 0; i < nchars; ++i) {
        const uint32_t chpos   = getUint32(&tmp[8*i]);
        const uint16_t chid    = mapCharacterId(encoding, getUint16(&tmp[8*i+4]));
        const uint16_t chwidth = getUint16(&tmp[8*i+6]);

        if (chwidth != 0) {
            // Read bitmap
            const uint32_t chsize = (chwidth+7)/8 * height;
            Array<char> bits(chsize);
            s.seek(base + chpos);
            s.readT(bits.ptr(), chsize);
            std::auto_ptr<GfxBitmapGlyph> glyph(new GfxBitmapGlyph(chwidth, height, (const uint8_t*) bits.ptr()));

            // Read anti-aliasing hints
            if (type == 3) {
                s.readT(header, 2);
                uint16_t nhints = getUint16(header);
                if (nhints != 0) {
                    Array<char> hints(4*nhints);
                    s.readT(hints.ptr(), 4*nhints);
                    for (uint32_t i = 0; i < nhints; ++i) {
                        glyph->addAAHint(getInt16(&hints[4*i]),
                                         getInt16(&hints[4*i+2]));
                    }
                }
            }

            addNewGlyph(chid, glyph.release());
        } else {
            addNewGlyph(chid, 0);
        }
    }
}
