/**
  *  \file util/modinit.cc
  *  \brief Module Initialisation
  *
  *  We'll be building several executables from PCC2 source code, each
  *  of which uses a different set of modules. To ensure that all
  *  modules are initialized correctly, and are notified of changes in
  *  data availability, modules can define a ModuleInitializer[]
  *  object that contains all initializer routines and conditions. The
  *  ModuleInitializer[] can contain any number of routines. It is
  *  terminated by a null entry.
  *
  *  The main program defines a null-terminated array of those
  *  ModuleInitializer*, and hands it to performInitialisation upon
  *  every state change. The flags specify the kind of state change. 
  *  Restrictions about the sequence of calls are listed in the header
  *  file. It is possible to perform multiple transitions in one call
  *  by specifying the bitwise or of multiple conditions.
  *
  *  As a simplification, programs that do not use any module other
  *  than mod_base_console, and do not use any other state than
  *  OnStartup do not define an array, but call the initializer
  *  directly.
  *
  *  The script "scripts/checkinit.pl" can be used to validate the
  *  ModuleInitializer infrastructure.
  *
  *  (c) 2001-2019 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "util/modinit.h"

/** Perform initialisation of a single module.
    \param flags State transitions to perform
    \param init  Module initializer, null terminated */
void
performInitialisation(uint32_t flags, const ModuleInitializer* init)
{
    while (init->initializer != 0) {
        if (init->condition & flags)
            init->initializer();
        ++init;
    }
}

/** Perform initialisation of a module list.
    \param flags State transitions to perform
    \param init  Module list, null terminated */
void
performInitialisation(uint32_t flags, const ModuleInitializer*const* init)
{
    while (*init) {
        performInitialisation(flags, *init);
        ++init;
    }
}
